/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package ZEN_IDL.be;

import java.io.*;
import java.util.*;
import util.*;

/*********************************************************
* Generates Config.java file
* This is the configuration file that the server side
* ORB uses to load different transports and protocols.
*
* @author Olav Haugan
* @author Tarana Damania
* @author David McKinnon
*********************************************************/
public class MQCConfigGenerator
{
    private MQCCodeGenHelper codeGenHelper; /* The code generator helper object */
    private MQCConfig config; /* The configuration options */
        
    /*********************************************************
    * Constructor. Initializes variables.
    * 
    *********************************************************/
    public MQCConfigGenerator(MQCConfig config)
    {
        codeGenHelper = new MQCCodeGenHelper(config);       
        this.config = config;
    }
    
    /*********************************************************
    * Generates the Config.java file
    *
    *********************************************************/
    public void generate()
    {
        PrintWriter pw = codeGenHelper.makeFile("", "Config");

        pw.println("package mqc;");

        pw.println("public class Config");
        pw.println("{");

        pw.print("\tpublic static final boolean DEBUG       = ");
        if (config.getDebugStatus()) {
            pw.println("true;");            
        } else {
            pw.println("false;");
        }
        pw.print("\tpublic static final int     DEBUG_LEVEL = ");
        pw.println(config.getDebugLevel() + ";");
        pw.println("");

        pw.print("\tpublic static final boolean DEBUG_MEMORY = ");
        if (config.getDebugMemoryStatus()) {
            pw.println("true;");            
        } else {
            pw.println("false;");
        }
        pw.print("\tpublic static final boolean DEBUG_TIMING = ");
        if (config.getDebugTimingStatus()) {
            pw.println("true;");            
        } else {
            pw.println("false;");
        }

        pw.println("\tpublic static final int GIOP = 0;");
        pw.println("\tpublic static final int GIOPLite = 1;");
        pw.println("\tpublic static final int MQCIOP = 2;");    
        pw.println("\tpublic static final int TCP = 0;");
        pw.println("\tpublic static final int UDP = 1;");
        pw.println("\tpublic static final int Go_Back_N_UDP = 2;");
        pw.println("\tpublic static final int Stop_N_Wait_UDP = 3;");
        pw.println("\tpublic static final int Serial = 4;");
        pw.println("\tpublic static final int OneWire = 5;");
        pw.println();
        pw.print("\tpublic static final int[] transports = {");

        Vector transports = config.getServerTransport();
        Enumeration e = transports.elements();
    
        while(e.hasMoreElements())
        {
            pw.print((String) e.nextElement());
            if(e.hasMoreElements())
            {
                pw.print(", ");
            }
        }
        pw.println("};");
        pw.print("\tpublic static final int[] protocols = {");

        Vector protocols = config.getServerProtocol();
        e = protocols.elements();
    
        while(e.hasMoreElements())
        {
            pw.print((String) e.nextElement());
            if(e.hasMoreElements())
            {
                pw.print(", ");
            }
        }
        pw.println("};");

        if(config.isForcingEndianess() > 0)
        {
            pw.println("\tpublic static final boolean FORCE_ENDIANESS = true;");            
            if(config.isForcingEndianess() == 1)
                pw.println("\tpublic static final boolean BIG_ENDIAN = true;");         
            else
                pw.println("\tpublic static final boolean BIG_ENDIAN = false;");            
        }
        else                
        {
            pw.println("\tpublic static final boolean FORCE_ENDIANESS = false;");           
            pw.println("\tpublic static final boolean BIG_ENDIAN = true;");         
        }
        
        // Check for temporal redundancy setting
        pw.println("\tpublic static final int temporal = " + config.getFTTemporal() + ";");
                
        // check to see if we need to include the GroupCom options
        Hashtable htGroupCom = config.getFTGroupCom();
        String strEnableGC = (String) htGroupCom.get("enableGroupCom");
        if(strEnableGC.equalsIgnoreCase("true")) {
            // here Group Communication has been enabled--load the config
            //  file with the configuration selections
            String strOrder = (String) htGroupCom.get("ftGCOrder");
            pw.println("\tpublic static final String GCOrder = \"" + strOrder + "\";");
            String strTimeOut = (String) htGroupCom.get("ftGCTimeOut");
            pw.println("\tpublic static final int GCTimeOut = " + strTimeOut + ";");
            String strTransport = (String) htGroupCom.get("ftGCTrans");
            pw.println("\tpublic static final String GCTransport = \"" + strTransport.toUpperCase() + "\";"); 
        }

        pw.print("\tpublic static final boolean CLDC = ");          
        if (config.getCLDCStatus())
        {
            pw.println("true;");
        }
        else
        {
            pw.println("false;");
        }
    
        // Set the value of the size of the packets to be transmitted. 
        // If intPacketSize == 0 it means that variable length packets will be
        // sent, so only check to make sure that the value is non-negative
        int intPacketSize = config.getMSGPayload();
        if(intPacketSize < 0) {
            intPacketSize = 0;
        }
        pw.println("\tpublic static final int intPacketSize = " + intPacketSize + ";");

        if (config.getSecurityAuditStatus()) {
            // only need the log length value when the Audit status is true
            pw.println("\tpublic static final int secAuditLogLen = " + config.getSecurityAuditLogLen());
        }
        
        pw.println("};");
        pw.println("// End of file.");
        pw.flush();
        pw.close();
    }
    
}
