/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

include(`Config_macros.m4')

package mqc;

import mqc.Config;
import mqc.POA;
import mqc.ior.*;
import mqc.protocols.*;
import mqc.holders.*;
import mqc.transports.*;
__GROUPCOM__(
import mqc.groupcom.*;
)
import java.io.*;
__J2SE__(import java.net.*;)
import java.util.*;

/**
 * Server ORB
 *
 * @author Olav Haugan
 * @author A. David McKinnon
 * @author Kim Christian Swenson
 * @author Ryan Johnston
 */
__GROUPCOM__(
public class S_ORB extends ORB implements MessageListener
)
__NOGROUPCOM__(
public class S_ORB extends ORB
)
{
   /* A list of POAs */
   private Hashtable servants;
   private Hashtable protocols;
   private Hashtable transports;

   /** Shutdown flag -- global for Server ORB */
   public static boolean keepRunning = true;

   /* a GroupCom object */
   __GCUNIFORM__(
   public GroupComUniform group;
   )
   __GCNONUNIFORM__(
   public GroupCom group;
   )

   /**
    * Constuctor. Initializes variables
    */
   public S_ORB()
   {
      servants = new Hashtable(1);
      protocols = new Hashtable(1);
      transports = new Hashtable(1);
      init();
   }

   /**
    * Converts an IOR string to an object.
    *
    * @param String An IOR string
    * @return Object The object
    */
   public Object string_to_object(String str)
   {
      IOR ior = new IOR(str);
      ior.decode();
      SDelegate delegate = new SDelegate(this, ior);
      SObjImpl objectImpl = new SObjImpl();
      objectImpl.set_delegate(delegate);
      return objectImpl;
   }

   /**
    * Converts a corbaloc string to an object.
    *
    * @param String A corbaloc string
    * @return Object The object
    */
   public Object corbaloc_to_object(String str)
   {
      IOR ior = new IOR(null, str);
      ior.decodeCorbaloc();
      SDelegate delegate = new SDelegate(this, ior);
      SObjImpl objectImpl = new SObjImpl();
      objectImpl.set_delegate(delegate);
      return objectImpl;
   }

   /**
    * Converts an object to a string
    *
    * @param Object The object
    * @return String An IOR string
    */
   public String object_to_string(Object obj)
   {
      SDelegate delegate = ((SObjImpl) obj).get_delegate();
      IOR ior = delegate.ior;
      ior.encode();
      return ior.getIORStr();
   }

   /**
    * Converts an object to a corbaloc URL string
    *
    * @param Object The object
    * @return String A corbaloc string
    */
   public String object_to_corbaloc(Object obj)
   {
      SDelegate delegate = ((SObjImpl) obj).get_delegate();
      IOR ior = delegate.ior;
      ior.encodeCorbaloc();
      return ior.getCorbalocStr();
   }

   /**
    * Adds a POA to the list of servants
    *
    * @param String Name of POA
    * @param POA The poa object itself
    */
   public void addPOA(String name, POA poa)
   {
      servants.put(name, poa);
   }

   /**
    * Finds the POA
    *
    * @param String Name of POA
    * @return POA The poa
    */
   public POA findPOA(String name)
   {
      return (POA) servants.get(name);
   }

   /**
    * Returns the transports
    *
    * @return Vector List of transports
    */
   public Hashtable getTransports()
   {
      return transports;
   }

   /**
    * Returns the protocols
    *
    * @return Vector List of protocols
    */
   public Hashtable getProtocols()
   {
      return protocols;
   }

   /**
    * Initializes protocol vectors and transport vectors
    */
   private void init()
   {
      int noTransports = Config.transports.length;
      int trans = 0;

      for(int i = 0; i < noTransports; i++)
      {
         trans = Config.transports[i];
         switch(trans)
         {
         __TCP_TP__(
         case Config.TCP:
            transports.put("TCP", new S_TCPTransport(this));
            break;
         )
         __UDP_TP__(
         case Config.UDP:
         case Config.Go_Back_N_UDP:
         case Config.Stop_N_Wait_UDP:
            transports.put("UDP", new S_UDPTransport(this));
            break;
         )
         __SERIAL_TP__(
         case Config.Serial:
            transports.put("Serial", new S_SerialTransport(this));
            break;
         )
         __ONEWIRE_TP__(
         case Config.OneWire:
            transports.put("1-wire", new S_OneWireTransport(this));
            break;
         )
         default:
            System.out.println("Unsupport transport in S_ORB");
            break;
         }
      }
   }

   /**
    * Starts up threads for each transport layer supported
    */
   public void run()
   {
      Enumeration e = transports.elements();
      S_Transport transport = null;

      while(e.hasMoreElements())
      {
         transport = (S_Transport) e.nextElement();
         Thread t = new Thread(transport);
         t.start();
      }
   }

   /**
    * Shut things down--somewhere/somehow someone wants to quit
    */
   public static void shutDown(int code)
   {
      System.exit(code);
   }

   /**
    * @return  request message instance of appropriate type
    */
   public Request findProtocolHandler(Object handler, byte[] buffer)
   {
      Request req;
      if ( (req = (Request) protocols.get(handler)) != null)
      {
         return req;
      }
      else if( (buffer[0] == GIOPHeader.GMAGIC) 
            && (buffer[1] == GIOPHeader.IMAGIC) 
            && (buffer[2] == GIOPHeader.OMAGIC) 
            && (buffer[3] == GIOPHeader.PMAGIC) )
      {
         __GIOP_PROTOCOL__(
         __GIOP10__( 
         req = new GIOPRequest(new GIOPHeader_1_0(), new GIOPRequestHeader_1_0()); 
         )
         __GIOP11__( 
         req = new GIOPRequest(new GIOPHeader_1_1(), new GIOPRequestHeader_1_1()); 
         )
         __GIOP12__( 
         req = new GIOPRequest(new GIOPHeader_1_2(), new GIOPRequestHeader_1_2());
         )
         __GIOP13__( 
         req = new GIOPRequest(new GIOPHeader_1_3(), new GIOPRequestHeader_1_3());
         )
         protocols.put(handler, req); 
         return req;
         )
      } else if( (buffer[0] == MQCIOPHeader.MMAGIC) 
            && (buffer[1] == MQCIOPHeader.QMAGIC)
            && (buffer[2] == MQCIOPHeader.CMAGIC) )
      {
         __MQCIOP_PROTOCOL__(
         req = new MQCIOPRequest();
         protocols.put(handler, req);
         return req;
         )
      }
      else
      {
         __GIOPLITE_PROTOCOL__(
         req = new GIOPLiteRequest();
         protocols.put(handler, req);
         return req;
         )
      }
      return null;
   }

   /**
    * @return success of attempting to remove handler
    */
   public boolean removeProtocolHandler(Object handler)
   {
      if ( protocols.remove(handler) != null)
      {
         System.out.println("Removed protocol handler for: " + handler);
         return true;
      }
      else
      {
         return false;
      }
   }

   /**
    * Handles the request from transport layer
    *
    * @param msg  a pre-deMarshalled message that is to be served
    */
   public Marshalable serveRequest( Request msg)
   {
      byte[] objKey = msg.getObjectKey();

      int poaLen = ((objKey[0] & 0xFF) << 24) |
                   ((objKey[1] & 0xFF) << 16) |
                   ((objKey[2] & 0xFF) << 8) |
                   (objKey[3] & 0xFF);

      byte[] poaName = new byte[poaLen];
      System.arraycopy(objKey, 4, poaName, 0, poaLen);

      POA poa = findPOA(new String(poaName));

      // Let the POA handle the request
      return poa.handleRequest( msg );
   }

   __GROUPCOM__(
   /**
    * Handles the request from Group Communications package
    *
    * @param ServerMessage The message that need to be handled
    */
   public void serveRequest(mqc.protocols.S_ServerMessage message)
   {

      byte[] objKey = message.getObjectKey();

      int bufferPtr = 0;

      int poaLen = ((objKey[bufferPtr++] & 0xFF) << 24) |
                   ((objKey[bufferPtr++] & 0xFF) << 16) |
                   ((objKey[bufferPtr++] & 0xFF) << 8) |
                   (objKey[bufferPtr++] & 0xFF);

      byte[] poaName = new byte[poaLen];

      System.arraycopy(objKey, bufferPtr, poaName, 0, poaLen);

      POA poa = findPOA(new String(poaName));
      Reply reply = poa.handleRequest(message);

      if(reply != null)
      {
         // For now do nothing with the replies eventually could send
         // replies back to group for consolidation of responses
         //System.out.println(reply.getBuffer()[20] + " " +  reply.size());
         //transport.send(reply.getBuffer());
      }
   }



   /**
    *define a method that is called whenever a MessageEvent takes place
    */
   public void receiveMessage(MessageEvent me)
   {
      // first grab a timestamp
      long lngCurTime = System.currentTimeMillis();
      // first we need to extract the vector containing the information about
      //  the message
      Vector vMessageInfo = me.getMessageData();
      // extract the message data from the vector
      String strSender = (String) vMessageInfo.elementAt(0);
      String strGroupList = (String) vMessageInfo.elementAt(1);
      String strMessageID = (String) vMessageInfo.elementAt(2);
      String strMessageData = (String) vMessageInfo.elementAt(3);
      String strTemp = (String) vMessageInfo.elementAt(4);
      long lngSendTime = Long.parseLong(strTemp);

      __DEBUG__(
      System.out.println("Received GroupCom message in S_ORB: " + strMessageData);
      )
      // next we need to process the CORBA request message
      // first convert the string back into a byte array
      byte[] bMessage;
      try
      {
         bMessage = strMessageData.getBytes();
         //__DEBUG__( System.out.println("Number of bytes read: " + bMessage.length);)
      }
      catch (Exception e)
      {
         __DEBUG__(
         e.printStackTrace();
         )
         return;
      }

      // next attach some input streams to the byte array for processing
      ByteArrayInputStream baiIn = new ByteArrayInputStream(bMessage);
      DataInputStream disIn = new DataInputStream(baiIn);

      byte[] buffer = new byte[4];

      try
      {
         disIn.readFully(buffer,0,4);
      }
      catch(Exception e)
      {
         return;
      }
      S_ServerMessage message = findProtocolHandler(buffer);
      //message = message.deMarshal(buffer, in);
      message = message.deMarshal(buffer, disIn);

      // next check the message request id
      int intReqID = message.getReqID();
      __DEBUG__(
      System.out.println("ReqID: " + intReqID);
      )

      // scan the buffer to see if this message has already been received
      //Integer intCReqID = new Integer(intReqID);
      //if(vMessageBuffer.contains(intCReqID)) {
      //  System.out.println("Dup.ReqID: " + intReqID);
      //  // here we need to ignore the message completely
      //} else {
      //  // here we need to insert the new id into the buffer
      //  System.out.println("New ReqID: " + intReqID);
      //  vMessageBuffer.addElement(intCReqID);
      //  if(vMessageBuffer.size() > 50) {
      //      vMessageBuffer.removeElementAt(0);
      //  }
      serveRequest(message);
      //}
   }

   // create a method to join an existing group
   public void joinGroup(String strServerInfo)
   {
      this.group.joinGroup(strServerInfo);
      try
      {
         Thread.sleep(1000);
      }
      catch (Exception e)
      {
         e.printStackTrace();
      }
   }

   )

   __GCUNIFORM__(
   // create a method to instanciate the groupcom object
   public void instanciateGroupCom(String strHostName, int intListenPort)
   {


      String strListenPort = Integer.toString(intListenPort);
      group = new GroupComUniform(intListenPort, strHostName + ":" +
                                  strListenPort, strHostName,
                                  Config.GCTransport, Config.GCOrder);
      group.startListener();
      try
      {
         Thread.sleep(1000);
      }
      catch (Exception e)
      {
         e.printStackTrace();
      }
      group.addMessageListener(this);

   }
   )
   __GCNONUNIFORM__(
   // create a method to instanciate the groupcom object
   public void instanciateGroupCom(String strHostName, int intListenPort)
   {

      String strListenPort = Integer.toString(intListenPort);
      group = new GroupCom(intListenPort, strHostName + ":" +
                           strListenPort, strHostName,
                           Config.GCTransport, Config.GCOrder);
      group.startListener();
      try
      {
         Thread.sleep(1000);
      }
      catch (Exception e)
      {
         e.printStackTrace();
      }
      group.addMessageListener(this);
   }
   )
}
