/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package mqc.protocols;

import mqc.Config;

/**
 * Hello.
 *
 * struct GIOPLiteRequestHeader {
 *    unsigned long      request_id;
 *    boolean            response_expected;
 *    sequence<octet>    object_key;
 *    string             operation;
 * };
 *
 */
public class GIOPLiteRequest implements GIOPLiteMessage, Request
{
   protected GIOPLiteHeader m_GIOPLiteHeader;

   /** The object key in what object to invoke the received operation on */
   protected byte[] m_objectKey;

   /** Which operation (method) to invoke on the object */
   protected String m_operation;

   /** An unique Request ID for the message (particulary needed for group communication) */
   protected int m_reqID;

   /** Indicates whether or not an reply message is expected */
   protected boolean m_oneWay;

   /** Size of request header (not included paddings, objKey and operation string. */
   public static final int SIZE_REQ_HDR = 16;

   protected Body m_body;

   /**
    * Constructor initializing variables..
    */
   public GIOPLiteRequest()
   {
      this( new GIOPLiteHeader( GIOPLiteMessage.REQUEST_MSG ) );
   }

   public GIOPLiteRequest( GIOPLiteHeader header )
   {
      m_GIOPLiteHeader = header;
      m_body = new Body();
   }

   public GIOPLiteHeader getGIOPLiteHeader()
   {
      return m_GIOPLiteHeader;
   }

   public void setGIOPLiteHeader( GIOPLiteHeader header )
   {
      m_GIOPLiteHeader = header;
   }

   public Body getBody()
   {
      return m_body;
   }

   public void setBody( Body body )
   {
      m_body = body;
   }

   public int getHdrSize()
   {
      return 8;
   }

   /**
    * Indicates whether or not a reply is required for this message (Request Messages Only)
    * @return Boolean indicating whether or no a reply is needed for this message
    */
   public boolean isOneWay()
   {
      return m_oneWay;
   }

   public void setOneWay( boolean oneway )
   {
      m_oneWay = oneway;
   }

   /**
    * Get the object key (Request Messages only!)
    * @return The Object Key
    */
   public byte [] getObjectKey()
   {
      return m_objectKey;
   }

   public void setObjectKey( byte[] key )
   {
      m_objectKey = key;
   }

   /**
    * Retrieves the operation to perform (Request Messages Only)
    * @return The current operation to invoke
    */
   public String getOperation()
   {
      return m_operation;
   }

   public void setOperation( String op )
   {
      m_operation = op;
   }

   /**
    * Returns the request ID from this message
    */
   public int getRequestID()
   {
      return m_reqID;
   }

   public void setRequestID( int id )
   {
      m_reqID = id;
   }

   public boolean isBigEndian()
   {
      return !(Config.FORCE_ENDIANESS && !Config.BIG_ENDIAN);
   }

   public void setBigEndian( boolean bigEndian )
   {
   }

   public void setMsgSize( byte[] buf, int size, boolean bigEndian )
   {
      GIOPLiteHeader.setMsgSize( buf, size, bigEndian );
   }

   public void marshal( CDROutputStream out )
   {
      out.reset();
      out.setBigEndian( isBigEndian() );

      // GIOPLite header
      m_GIOPLiteHeader.marshal( out );

      // unsigned long request_id
      out.writeUnsignedLong( m_reqID );

      // boolean response_expected
      out.writeBoolean( ! m_oneWay );

      // sequence<octet> object_key
      out.writeOctetSequence( m_objectKey );

      // string operation
      out.writeString( m_operation );

      // GIOPLite Request Body

      // Copy data from body buffer into underlying
      // buffer of CDROutputStream, and adjust
      // the Stream's pointer acordingly.

      // get underlying buffer
      byte[] buf = out.getBuffer();
      
      // Get bufferPointer and align it with 4-bytes
      int bufPtr = out.bytesStreamed();
      bufPtr = (bufPtr + 3) & ~3;

      // Hack ... set message size again now that we know it .. !
      m_GIOPLiteHeader.setSize( bufPtr - getHdrSize() + m_body.getPtr() );
      GIOPLiteHeader.setMsgSize( buf, m_GIOPLiteHeader.getSize(), isBigEndian() );

      // Resize if necessary
      if ( buf.length < bufPtr + m_body.getPtr() )
      {
         byte[] tmp = new byte[bufPtr + m_body.getPtr()];
         System.arraycopy(buf, 0, tmp, 0, out.bytesStreamed() );
         buf = tmp;
      }

      // copy data
      System.arraycopy( m_body.getBuffer(), 0, buf, bufPtr, m_body.getPtr() );

      // Insert underlying buffer in case of resize
      out.setBuffer( buf );

      // Adjust buffer pointer
      out.setPtr( bufPtr + m_body.getPtr() );
   }

   public void deMarshal( CDRInputStream in ) throws RuntimeException
   {
      in.reset();
      in.setBigEndian( isBigEndian() );

      // GIOPLite header
      m_GIOPLiteHeader.deMarshal( in );

      // unsigned long request_id
      m_reqID = in.readUnsignedLong( );

      // boolean response_expected
      m_oneWay = ! in.readBoolean( );

      // sequence<octet> object_key
      m_objectKey = in.readOctetSequence( );

      // string operation
      m_operation = in.readString( );

      // GIOPLite Request Body

      // Copy data from underlying buffer of 
      // CDRInputStream into body buffer, and adjust
      // the Stream's pointer acordingly.

      // Get bufferPointer and align it with 4-bytes
      int bufPtr = in.bytesStreamed();
      bufPtr = (bufPtr + 3) & ~3;

      // Make sure message body's buffer is big enough
      m_body.setMinSize( m_GIOPLiteHeader.getSize() - bufPtr + getHdrSize() );

      // copy body part of message from inputstream into body
      System.arraycopy( in.getBuffer(), bufPtr, m_body.getBuffer(), 0, m_GIOPLiteHeader.getSize() - bufPtr + getHdrSize() );

      // Adjust buffer pointer
      in.setPtr( m_GIOPLiteHeader.getSize() + getHdrSize() );
   }
}
