/*
 * @(#)SecureRandom.java    1.41 01/12/03
 *
 * Copyright 2002 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package mqc.security;

import java.util.Enumeration;

/**
 * <p>This class provides a cryptographically strong pseudo-random number
 * generator (PRNG). A cryptographically strong pseudo-random number
 * minimally complies with the statistical random number generator tests
 * specified in <a href="http://csrc.nist.gov/cryptval/140-2.htm"><i>FIPS 140-2, Security Requirements for Cryptographic Modules</i></a>, section 4.9.1.
 * Additionally, SecureRandom must produce non-deterministic 
 * output and therefore it is required that the seed material be unpredictable
 * and that output of SecureRandom be cryptographically strong sequences as
 * described in <a href="http://www.ietf.org/rfc/rfc1750.txt"><i>RFC 1750: Randomness Recommendations for Security</i></a>.
 * 
 * <p>Like other algorithm-based classes in Java Security, SecureRandom 
 * provides implementation-independent algorithms, whereby a caller 
 * (application code) requests a particular PRNG algorithm
 * and is handed back a SecureRandom object for that algorithm. It is
 * also possible, if desired, to request a particular algorithm from a
 * particular provider. See the <code>getInstance</code> methods.
 *
 * <p>Thus, there are two ways to request a SecureRandom object: by
 * specifying either just an algorithm name, or both an algorithm name
 * and a package provider.
 *
 * <ul>
 *
 * <li>If just an algorithm name is specified, as in:
 * <pre>
 *      SecureRandom random = SecureRandom.getInstance("SHA1PRNG");
 * </pre>
 * the system will determine if there is an implementation of the algorithm
 * requested available in the environment, and if there is more than one, if
 * there is a preferred one.<p>
 * 
 * <li>If both an algorithm name and a package provider are specified, as in:
 * <pre>
 *      SecureRandom random = SecureRandom.getInstance("SHA1PRNG", "SUN");
 * </pre>
 * the system will determine if there is an implementation of the
 * algorithm in the package requested, and throw an exception if there
 * is not.
 *
 * </ul>
 *
 * <p>The SecureRandom implementation attempts to completely
 * randomize the internal state of the generator itself unless
 * the caller follows the call to a <code>getInstance</code> method
 * with a call to the <code>setSeed</code> method:
 * <pre>
 *      SecureRandom random = SecureRandom.getInstance("SHA1PRNG");
 *      random.setSeed(seed);
 * </pre>
 *
 * <p>After the caller obtains the SecureRandom object from the
 * <code>getInstance</code> call, it can call <code>nextBytes</code>
 * to generate random bytes:
 * <pre>
 *      byte bytes[] = new byte[20];
 *      random.nextBytes(bytes);
 * </pre>
 *
 * <p>The caller may also invoke the <code>generateSeed</code> method
 * to generate a given number of seed bytes (to seed other random number
 * generators, for example):
 * <pre>
 *      byte seed[] = random.generateSeed(20);
 * </pre>
 *
 * @see java.security.SecureRandomSpi
 * @see java.util.Random
 * 
 * @version 1.41, 12/03/01
 * @author Benjamin Renaud
 * @author Josh Bloch 
 */

public class SecureRandom extends java.util.Random
{

   // Seed Generator
   private static SecureRandom seedGenerator = null;

   /**
    * <p>By using this constructor, the caller obtains a SecureRandom object
    * containing the implementation from the highest-priority installed
    * provider that has a SecureRandom implementation.
    * 
    * <p>Note that this instance of SecureRandom has not been seeded.
    * A call to the <code>setSeed</code> method will seed the SecureRandom
    * object.  If a call is not made to <code>setSeed</code>, the first call
    * to the <code>nextBytes</code> method will force the SecureRandom object
    * to seed itself.
    *
    * <p>This constructor is provided for backwards compatibility. 
    * The caller is encouraged to use one of the alternative
    * <code>getInstance</code> methods to obtain a SecureRandom object.
    */
   public SecureRandom()
   {
      /*
       * This call to our superclass constructor will result in a call
       * to our own <code>setSeed</code> method, which will return
       * immediately when it is passed zero.
       */
      super(0);
   }

   /**
    * <p>By using this constructor, the caller obtains a SecureRandom object
    * containing the implementation from the highest-priority installed
    * provider that has a SecureRandom implementation. This constructor 
    * uses a user-provided seed in
    * preference to the self-seeding algorithm referred to in the empty
    * constructor description. It may be preferable to the empty constructor
    * if the caller has access to high-quality random bytes from some physical
    * device (for example, a radiation detector or a noisy diode).
    * 
    * <p>This constructor is provided for backwards compatibility. 
    * The caller is encouraged to use one of the alternative
    * <code>getInstance</code> methods to obtain a SecureRandom object, and
    * then to call the <code>setSeed</code> method to seed it.
    * 
    * @param seed the seed.
    */
   public SecureRandom(byte seed[])
   {
      super(0);
   }


   // Declare serialVersionUID to be compatible with JDK1.1
   static final long serialVersionUID = 4940670005562187L;

   private long counter;
}
