/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

//include(`Config_macros.m4')
/**
 * BlockCipher.java.m4
 *
 * This file started out as cryptix.jce.provider.cipher.BlockCipher. It was
 * modified to fit into the MQC framework. It requires the use of macros
 in order to implement cipher modes (ECB, CBC) and padding (NONE, PKCS5).
 *----
 * Id: BlockCipher.java,v 1.14 2000/08/03 13:21:57 pw Exp
 *
 * Copyright (C) 1995-2000 The Cryptix Foundation Limited.
 * All rights reserved.
 *
 * Use, modification, copying and distribution of this software is subject
 * the terms and conditions of the Cryptix General Licence. You should have
 * received a copy of the Cryptix General Licence along with this library;
 * if not, you can download a copy from http://www.cryptix.org/ .
 * 
 * @author  Jeroen C. van Gelderen (gelderen@cryptix.org)
 * @author  Paul Waserbrot (pw@cryptix.org)
 *  version Revision: 1.14 
 * @author  David McKinnon
 */
package mqc.security.cipher;

//import java.security.AlgorithmParameters;
//import java.security.InvalidAlgorithmParameterException;
import mqc.security.InvalidKeyException;
import mqc.security.Key;
import mqc.security.Cipher;
import mqc.security.NoSuchAlgorithmException;
//import java.security.NoSuchProviderException;
import mqc.security.SecureRandom;
//import java.security.spec.AlgorithmParameterSpec;
//import java.security.spec.InvalidParameterSpecException;
import mqc.security.BadPaddingException;
import mqc.security.Cipher;
//import javax.crypto.CipherSpi;
import mqc.security.IllegalBlockSizeException;
import mqc.security.NoSuchPaddingException;

import mqc.security.SecureRandom;
import mqc.security.GeneralSecurityException;
import mqc.security.ShortBufferException;



/**
 * BlockCipher extends Cipher. It relies upon MQC macros to enable
 * parameterized functionality. (A weakness on the server side, but acceptable
 * on the client-side).
 */
public abstract class BlockCipher
   extends Cipher
{

   // Variables and constants
   //............................................................................

   /** State */
   private int state = STATE_UNINITIALIZED;
   private static final int
      STATE_UNINITIALIZED = 0,
      STATE_DECRYPT       = 1,
      STATE_ENCRYPT       = 2;


   /** Our blocksize */
   private final int BLOCK_SIZE;


   /** Our key. We need it when we reset ourselves. */
   private Key key;


   /** What algorithm is used */
   private String algorithm;


   // Cipher Mode stuff
   private int          bufCount;
   private final byte[] buf;

   __SEC_CIPHER_MODE_CBC__(
   /** previous ciphertext block (during decryption only) */
   private final byte[] prevBlock;
   /** IV used for CBC */
   private final byte[] IV;
   )

   // Constructors and Object() methods
   //............................................................................


   /**
    * Construct a new BlockCipher (CipherSpi) with the given block size.
    */
   protected BlockCipher(int blockSize)
   {
      super("__SEC_CIPHER_ALG__/__SEC_CIPHER_MODE__/__SEC_CIPHER_PADDING__");

      BLOCK_SIZE = blockSize;

      state     = STATE_UNINITIALIZED;
      bufCount  = 0;
      buf       = new byte[BLOCK_SIZE];

      __SEC_CIPHER_MODE_CBC__(
      prevBlock = new byte[BLOCK_SIZE];
      IV        = new byte[BLOCK_SIZE];
      )
   }


   /**
    * Always throws a CloneNotSupportedException (cloning of ciphers is not
    * supported for security reasons).
    */
   public final Object clone() throws CloneNotSupportedException
   {
      throw new CloneNotSupportedException();
   }


   // CipherSPI implementation
   //............................................................................
   // mode set by macros
   protected final void engineSetMode(String mode)
      throws NoSuchAlgorithmException
   {
      // mode set by macros, complain if mode does not match
      if ( !mode.toUpperCase().equals("__SEC_CIPHER_MODE__"))
      {
      throw new NoSuchAlgorithmException ("Cipher mode <" + mode +
                                          "> is not (currently) supported");
      }
   }

   // padding set by macros
   protected final void engineSetPadding(String padding)
      throws NoSuchPaddingException
   {
      // padding set by macros, complain if padding does not match
      if ( !padding.toUpperCase().equals("__SEC_CIPHER_PADDING__"))
         throw new NoSuchPaddingException ("Cipher padding <" + padding +
                                           "> is not (currently) supported");
   }


   protected final int engineGetBlockSize()
   {
      return BLOCK_SIZE;
   }


   /**
    * Returns the length in bytes that an output buffer would need to be in
    * order to hold the result of the next update or doFinal operation, given
    * the input length <code>inputLen</code> (in bytes).
    *
    * This call takes into account any unprocessed (buffered) data from a
    * previous update call(s), and padding.
    *
    * The actual output length of the next <code>update or doFinal</code> call
    * may be smaller than the length returned by this method. For ciphers with
    * a padding, calling the update method will generally return less data
    * than predicted by this function.
    *
    * @param  inputLen the length in bytes.
    *
    * @return the maximum amount of data that the cipher will return.
    */
   protected final int engineGetOutputSize(int inputLen)
   {
      __SEC_CIPHER_PADDING_NONE__(
      return inputLen + bufCount;
      )
      __SEC_CIPHER_PADDING_PKCS5__(
      return inputLen + bufCount + engineGetPadSize(inputLen);
      )
   }


   /**
    * Returns a copy of the initialization vector (IV) used in this cipher.
    *
    * @return A copy of the IV  or null if this cipher does not have an IV or
    *         null if the IV has not yet been set.
    */
   protected final byte[] engineGetIV()
   {
      __SEC_CIPHER_MODE_ECB__(
      return null;
      )
      __SEC_CIPHER_MODE_CBC__(
      return IV;
      )
   }


   /**
    * Returns the padding size used by this cipher.
    *
    * @param  inputLen length of the input string
    *
    * @return the numbers of padding bytes required
    */
   protected final int engineGetPadSize(int inputLen)
   {
      __SEC_CIPHER_PADDING_NONE__(
      return 0;
      )
      __SEC_CIPHER_PADDING_PKCS5__(
      return BLOCK_SIZE - (inputLen + bufCount) % BLOCK_SIZE;
      )
   }


   /**
    * Initialize this blockcipher for encryption or decryption.
    *
    * If the cipher requires randomness, it is taken from <code>random</code>.
    * Randomness is required for modes that use IVs and might be required for
    * some padding schemes.
    *
    * @param opmode Cipher.ENCRYPT_MODE or Cipher.DECRYPT_MODE.
    * @param key    secret key
    * @param random source of randomness
    */
   public final void engineInit(int opmode, Key key, SecureRandom random)
   throws InvalidKeyException
   {
      //-- Init the Padding stuff --
      // nothing to do for NONE or PKCS5 (or any other padding type)

      // FIXME: put in BlockCipher mode init code here
      //-- Init the Mode stuff --
      __SEC_CIPHER_MODE_ECB__(
      // nothing to do
      )
      __SEC_CIPHER_MODE_CBC__(
      //FIXME: FIXME: Put in a better IV, perhaps rework the init routines to include it
      IV[0] = 'm';
      IV[1] = 'q';
      IV[2] = 'c';
      )

      switch (opmode)
      {
      case Cipher.ENCRYPT_MODE:
         coreInit( key, false);
         __SEC_CIPHER_MODE_CBC__(
         System.arraycopy(IV, 0, buf, 0, BLOCK_SIZE);
         )
         state = STATE_ENCRYPT;
         break;
      case Cipher.DECRYPT_MODE:
         coreInit( key, true);
         __SEC_CIPHER_MODE_CBC__(
         System.arraycopy(IV, 0, prevBlock, 0, BLOCK_SIZE);
         )
         state = STATE_DECRYPT;
         break;
      default:
         throw new InvalidKeyException("Unreachable code reached");
      }

      bufCount = 0;       // reset the buffer
   }


   protected final int
   engineUpdate(byte[] input, int inputOffset, int inputLen,
                byte[] output, int outputOffset)
      throws ShortBufferException
   {
      throw new ShortBufferException("Called BlockCipher::engineUpdate--use doFinal");
      //        if (inputLen == 0)
      //           return 0;
      //return padding.update(input, inputOffset, inputLen, output, outputOffset);

      // FIXME: Add update code, for now do nothing
      //        return 0;
   }



   /** Implemented in terms of engineUpdate(byte[], int, int, byte[], int) */
   protected final byte[]
   engineUpdate(byte[] input, int inputOffset, int inputLen)
   {
      System.out.println("Called BlockCipher::engineUpdate--use doFinal");
      //throw new Exception("Called BlockCipher::engineUpdate--use doFinal");
      if (inputLen == 0)
         return null;
      try
      {
         byte[] tmp  = new byte[this.engineGetOutputSize(inputLen)];
         int i = this.engineUpdate(input, inputOffset, inputLen, tmp, 0);
         if (i != tmp.length)
         {
            byte [] t = new byte[i];
            System.arraycopy(tmp, 0, t, 0, i);
            tmp = t;
         }
         return tmp;
      }
      catch(ShortBufferException e)
      {
         throw new RuntimeException("PANIC: Unreachable code reached.");
      }
   }


   /**
    * @throws BadPaddingException
    *         (decryption only) if padding is expected but not found at the
    *         end of the data or the padding is found but corrupt
    * @throws IllegalBlockSizeException
    *         if no padding is specified and the input data is not a multiple
    *         of the blocksize.
    * @throws ShortBufferException if the given buffer is to short to hold
    *         the result.
    */
   public final int
   engineDoFinal(byte[] input, int inputOffset, int inputLen,
                 byte[] output, int outputOffset)
      throws ShortBufferException, IllegalBlockSizeException, BadPaddingException
   {
      __DEBUG__(if (state == STATE_ENCRYPT) System.out.println(">>ENCRYPT:");)
      __DEBUG__(if (state == STATE_DECRYPT) System.out.println("<<DECRYPT:");)
      __DEBUG__(System.out.println("engineDoFinal(input,"+inputOffset+","+inputLen+",output,"+outputOffset+")");)
      int saved_outputOffset = outputOffset;
      
		//-- First, the Encryption Padding
      byte [] inputPad = input;
      if (state == STATE_ENCRYPT)
      {
         __SEC_CIPHER_PADDING_NONE__(
         if (inputLen % BLOCK_SIZE != 0)
         {
            throw new BadPaddingException("inputLen = " + inputLen);
         }
         //inputPad = input; //<--not needed
         )
         __SEC_CIPHER_PADDING_PKCS5__(
			//number of bytes to encrypt
			
         int pad = engineGetPadSize(inputLen);
         __DEBUG__(System.out.println("pad= "+pad);)
			//Create a new buffer with the expected length
         inputPad = new byte[pad + inputLen];
			System.arraycopy(input, inputOffset, inputPad, 0, inputLen);
			//inputLen is used to count down the number of rounds to encrypt over the buffer
			for (int i = 0;  i < pad;  i++)
         {
				//write something over the last paddings...
            inputPad[inputLen + i] = (byte) pad;
         }
         inputLen += pad; //end of buffer (the number of bytes to encrypt)
			inputOffset = 0; //make sure that we are writing ove the correct bytes
         __DEBUG__(
         System.out.println("pad= " + pad + ", inputLen= " + inputLen);)
         )
      }

      //-- Second, the Mode
      int ret = 0;
      int remainder;

      __SEC_CIPHER_MODE_ECB__(
      while(inputLen >= (remainder = BLOCK_SIZE - bufCount))
      {
         System.arraycopy(inputPad, inputOffset, buf, bufCount, remainder);
         coreCrypt(buf, 0, output, outputOffset);
         inputLen     -= remainder;
         inputOffset  += remainder;
         outputOffset += BLOCK_SIZE;
         ret          += BLOCK_SIZE;
         bufCount      = 0;
         __DEBUG__(
         System.out.println("inputLen= " + inputLen + ", inputOffset= " + inputOffset + ", ret= " + ret);
         )
      }
      // Invariant: bufCount < CIPHER_BLOCK_SIZE bytes
      System.arraycopy(inputPad, inputOffset, buf, bufCount, inputLen);
      bufCount += inputLen;
      // Invariant: bufCount < CIPHER_BLOCK_SIZE bytes
      )

      __SEC_CIPHER_MODE_CBC__(
      if(state == STATE_DECRYPT)
      {
         while(inputLen >= (remainder = BLOCK_SIZE - bufCount))
         {
            for( int i=0; i<remainder; i++)
               buf[bufCount++] = inputPad[inputOffset++];
            // encrypt to output
            coreCrypt(buf, 0, output, outputOffset);
            // xor in previous block
            for( int i=0; i<BLOCK_SIZE; i++)
               output[outputOffset++] ^= prevBlock[i];
            // store cipher text as IV for next block
            for( int i=0; i<BLOCK_SIZE; i++)
               prevBlock[i] = buf[i];
            inputLen     -= BLOCK_SIZE;
            ret          += BLOCK_SIZE;
            bufCount      = 0;
         }
         for( int i=0; i<inputLen; i++)
            buf[bufCount++] = inputPad[inputOffset++];
      }
      else
      {
         while(inputLen >= (remainder = BLOCK_SIZE - bufCount))
         {
            for( int i=0; i<remainder; i++)
               buf[bufCount++] ^= inputPad[inputOffset++];
            coreCrypt(buf, 0, buf, 0);
            System.arraycopy(buf, 0, output, outputOffset, BLOCK_SIZE);
            inputLen     -= remainder;
            outputOffset += BLOCK_SIZE;
            ret          += BLOCK_SIZE;
            bufCount      = 0;
         }
         for( int i=0; i<inputLen; i++)
            buf[bufCount++] ^= inputPad[inputOffset++];
      }
      )

      //-- Third, the Decryption Padding
      if (state == STATE_DECRYPT)
      {
      __SEC_CIPHER_PADDING_NONE__(
      // no padding to strip
      )
      __SEC_CIPHER_PADDING_PKCS5__(
      ret -= output[saved_outputOffset + ret - 1];
      )
      }

      __DEBUG__(System.out.println("done, ret= " + ret);)
      return ret;
   }


   /**
    * Implemented in terms of engineDoFinal(byte[], int, int, byte[], int)
    *
    * @throws BadPaddingException
    *         (decryption only) if padding is expected but not found at the
    *         end of the data.
    * @throws IllegalBlockSizeException
    *         if no padding is specified and the input data is not a multiple
    *         of the blocksize.
    */
   public final byte[]
   engineDoFinal(byte[] input, int inputOffset, int inputLen)
      throws IllegalBlockSizeException, BadPaddingException
   {
      __DEBUG__(
      System.out.println("engineDoFinal(input,"+inputOffset+","+inputLen+")");
      )
      try
      {
         byte[] tmp  = new byte[this.engineGetOutputSize(inputLen)];
         int i = this.engineDoFinal(input, inputOffset, inputLen, tmp, 0);
         if (i != tmp.length)
         {
            byte [] t = new byte[i];
            System.arraycopy(tmp, 0, t, 0, i);
            tmp = t;
         }

         return tmp;
      }
      catch(ShortBufferException e)
      {
         throw new RuntimeException("PANIC: Unreachable code reached.");
      }
   }



   // Abstract BPI methods
   //
   // The abstract methods are overridden by the actual cipher implementations
   // The core* methods are called from within the Mode* classes
   //.............................................................................

   abstract void coreInit(Key key, boolean decrypt)
   throws InvalidKeyException;


   /**
    * Encrypt a given buffer. in and out can point to the same buffer if
    *   (outOffset == inOffset) || (outOffset >= (inOffset+coreGetBlockSize))
    * That is: the buffers may not partially overlap...
    */
   abstract void coreCrypt(byte[] in, int inOffset, byte[] out, int outOffset);


   int coreGetBlockSize()
   {
      return BLOCK_SIZE;
   }
}
